# Copyright 2020 StackHut Limited (trading as Datapane)
# SPDX-License-Identifier: Apache-2.0
from .ar_types import (  # noqa: F401
    ARROW_EXT,
    ARROW_MIMETYPE,
    HTML,
    JSON,
    MIME,
    PKL_MIMETYPE,
    SECS_1_HOUR,
    SECS_1_WEEK,
    SIZE_1_MB,
    TD_1_DAY,
    TD_1_HOUR,
    URL,
    EnumType,
    Hash,
    JDict,
    JList,
    NPath,
    SDict,
    SList,
    SSDict,
    StrEnum,
)
from .datafiles import ArrowFormat  # noqa: F401
from .utils import (  # noqa: F401
    guess_type,
    timestamp,
    unixtime,
    utf_read_text,
)
from .viewxml_utils import ViewXML, load_doc, validate_view_doc  # noqa: F401



import secrets

from .base import BaseBlock, BlockId


def gen_name() -> str:
    """Return a (safe) name for use in a Block"""
    return f"id-{secrets.token_urlsafe(8)}"


class Empty(BaseBlock):
    """
    An empty block that can be updated / replaced later

    Args:
        name: A unique name for the block to reference when updating the report
    """

    _tag = "Empty"

    def __init__(self, name: BlockId):
        super().__init__(name=name)



import os

from . import __version__
from .versioning import convert2semver

AR_VERSION = os.environ.get("AR_VERSION", __version__).removeprefix("v")
AR_CDN_BASE = os.environ.get(
    "AR_CDN_BASE",
    f"https://cdn.jsdelivr.net/npm/arakawa@{convert2semver(AR_VERSION)}/dist",
)



from __future__ import annotations

import dataclasses
from enum import Enum
from pathlib import Path
from typing import Generic, TypeVar

from arakawa.common import ViewXML
from arakawa.view import Blocks

from .file_store import DummyFileEntry, FileEntry, FileStore


@dataclasses.dataclass
class ViewState:
    # maybe a FileHandler interface??
    blocks: Blocks
    file_entry_klass: dataclasses.InitVar[type[FileEntry]]
    store: FileStore = dataclasses.field(init=False)
    view_xml: ViewXML = ""
    entries: dict[str, str] = dataclasses.field(default_factory=dict)
    dir_path: dataclasses.InitVar[Path | None] = None

    def __post_init__(self, file_entry_klass, dir_path):
        # TODO - should we use a lambda for file_entry_klass with dir_path captured?
        self.store = FileStore(fw_klass=file_entry_klass, assets_dir=dir_path)


P_IN = TypeVar("P_IN")
P_OUT = TypeVar("P_OUT")


class BaseProcessor(Generic[P_IN, P_OUT]):
    """Processor class that handles pipeline operations"""

    s: ViewState

    def __call__(self, x: P_IN) -> P_OUT:
        raise NotImplementedError("Implement in subclass")


# TODO - type this properly
class Pipeline(Generic[P_IN]):
    """
    A simple, programmable, eagerly-evaluated, pipeline specialized on ViewAST transformations
    similar to f :: State s => s ViewState x -> s ViewState y
    """

    # NOTE - toolz has an untyped function for this

    def __init__(self, s: ViewState, x: P_IN = None):
        self._state = s
        self._x = x

    def pipe(self, p: BaseProcessor[P_IN, P_OUT]) -> Pipeline[P_OUT]:
        p.s = self._state
        y = p.__call__(self._x)  # need to call as positional args
        self._state = p.s
        return Pipeline(self._state, y)

    @property
    def state(self) -> ViewState:
        return self._state

    @property
    def result(self) -> P_IN:
        return self._x


def mk_null_pipe(blocks: Blocks) -> Pipeline[None]:
    s = ViewState(blocks, file_entry_klass=DummyFileEntry)
    return Pipeline(s)


# Top-level API options / types
class Width(Enum):
    NARROW = "narrow"
    MEDIUM = "medium"
    FULL = "full"

    def to_css(self) -> str:
        if self == self.NARROW:
            return "max-w-3xl"
        if self == self.MEDIUM:
            return "max-w-screen-xl"
        return "max-w-full"


class TextAlignment(Enum):
    JUSTIFY = "justify"
    LEFT = "left"
    RIGHT = "right"
    CENTER = "center"


class FontChoice(Enum):
    DEFAULT = "Inter, ui-sans-serif, system-ui"
    SANS = "ui-sans-serif, sans-serif, system-ui"
    SERIF = "ui-serif, serif, system-ui"
    MONOSPACE = "ui-monospace, monospace, system-ui"


# Currently unused
# class PageLayout(Enum):
#     TOP = "top"
#     SIDE = "side"


@dataclasses.dataclass
class Formatting:
    """Configure styling and formatting"""

    bg_color: str = "#FFF"
    accent_color: str = "#4E46E5"
    font: FontChoice | str = FontChoice.DEFAULT
    text_alignment: TextAlignment = TextAlignment.LEFT
    width: Width = Width.MEDIUM
    light_prose: bool = False

    def to_css(self) -> str:
        font = self.font.value if isinstance(self.font, FontChoice) else self.font

        return f""":root {{
    --ar-accent-color: {self.accent_color};
    --ar-bg-color: {self.bg_color};
    --ar-text-align: {self.text_alignment.value};
    --ar-font-family: {font};
}}"""



"""
Arakawa Blocks API

Describes the collection of `Block` objects that can be combined together to make a `arakawa.utils.api.report.core.Report`.
"""

from __future__ import annotations

import sys
from abc import ABC
from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from lxml.builder import ElementMaker

from arakawa.common.viewxml_utils import is_valid_id, mk_attribs
from arakawa.exceptions import ARError
from arakawa.utils import log

if sys.version_info <= (3, 11):
    from typing_extensions import Self
else:
    from typing import Self

if TYPE_CHECKING:
    from arakawa.blocks import Block
    from arakawa.view import ViewVisitor

E = ElementMaker()  # XML Tag Factory


BlockId = str

VV = TypeVar("VV", bound="ViewVisitor")


class BaseBlock(ABC):
    """Base Block class - subclassed by all Block types

    ..note:: The class is not used directly.
    """

    _tag: str

    def __init__(self, name: BlockId | None = None, **kwargs: Any):
        """
        Args:
            name: A unique name to reference the block, used when referencing blocks via the report editor and when embedding
        """
        self._block_name: str = self._tag.lower()
        self.name = name

        # validate name
        if name and not is_valid_id(name):
            raise ARError(f"Invalid name '{name}' for block")

        self._attributes: dict[str, str] = {}
        self._add_attributes(name=name, **kwargs)

        self._truncate_strings(kwargs, "caption", 512)
        self._truncate_strings(kwargs, "label", 256)

    @staticmethod
    def _truncate_strings(kwargs: dict, key: str, max_length: int):
        if key in kwargs:
            x: str = kwargs[key]
            if x and len(x) > max_length:
                kwargs[key] = f"{x[:max_length-3]}..."
                log.warning(f"{key} currently '{x}'")
                log.warning(
                    f"{key} must be less than {max_length} characters, truncating"
                )
                # raise DPClientError(f"{key} must be less than {max_length} characters, '{x}'")

    def _add_attributes(self, **kwargs):
        self._attributes.update(mk_attribs(**kwargs))

    def _ipython_display_(self):
        """Display the block as a side effect within a Jupyter notebook"""
        from IPython.display import HTML, display

        from arakawa.ipython.environment import get_environment
        from arakawa.processors.api import stringify_report
        from arakawa.view import Blocks

        if get_environment().support_rich_display:
            html_str = stringify_report(Blocks(self))
            display(HTML(html_str))
        else:
            display(self.__str__())

    def accept(self, visitor: VV) -> VV:
        visitor.visit(self)
        return visitor

    def __str__(self) -> str:
        return f"<{self._tag} attribs={self._attributes}>"

    def __copy__(self) -> Self:
        """custom copy that deep copies attributes"""
        inst = self.__class__.__new__(self.__class__)
        inst.__dict__.update(self.__dict__)
        inst._attributes = self._attributes.copy()

        return inst


class DataBlock(BaseBlock):
    """Abstract block that represents a leaf-node in the tree, e.g. a Plot or Table

    ..note:: This class is not used directly.
    """


BlockOrPrimitive = Union[BaseBlock, Any]  # TODO - expand
BlockList = list[BaseBlock]


def wrap_block(b: BlockOrPrimitive) -> Block:
    from .wrappers import convert_to_block

    # if isinstance(b, Page):
    #     raise DPClientError("Page objects can only be at the top-level")
    if not isinstance(b, BaseBlock):
        # import here as a very slow module due to nested imports
        # from ..files import convert

        return convert_to_block(b)
    return cast("Block", b)



import abc
import dataclasses as dc
from collections.abc import Generator
from contextlib import contextmanager
from copy import copy
from typing import TYPE_CHECKING

from multimethod import multimethod

from arakawa import blocks as bk
from arakawa.blocks import BaseBlock
from arakawa.blocks.layout import ContainerBlock
from arakawa.exceptions import ARError
from arakawa.utils import log

from .view_blocks import Blocks

if TYPE_CHECKING:
    from arakawa.blocks.base import VV


@dc.dataclass
class ViewVisitor(abc.ABC):  # noqa: B024
    def visit(self: "VV", _: BaseBlock) -> "VV":
        return self


@dc.dataclass
class PrettyPrinter(ViewVisitor):
    """Print out the view in an indented, XML-like tree form"""

    indent: int = 1

    @multimethod
    def visit(self, b: BaseBlock):
        pass

    @visit.register  # type: ignore
    def _(self, b: ContainerBlock):
        self.indent += 2
        _ = b.traverse(self)
        self.indent -= 2


# TODO - split out into BlockBuilder helper here
@dc.dataclass
class PreProcess(ViewVisitor):
    """Block-level preprocessor, operations include,
    - Inline consecutive unnamed Text blocks
    - auto-column processing (TODO)

    We may need multiple passes here to make things easier

    We also transform the AST during the XMLBuilder visitor, and post-XML conversion in the ConvertXML processor.
    Where the transform should be placed is still under consideration, and depends on the nature of the transform.
    Some may be easier as a simple XSLT transform, others as more complex Python code operation either on the Blocks AST
    or the XML DOM via lxml.
    """

    in_collapsible_group: bool = False
    # is_finalized determines if we allow dynamic blocks or validate based on their static/current subblocks
    is_finalized: bool = True
    current_state: list[BaseBlock] = dc.field(default_factory=list)
    current_text: list[bk.Text] = dc.field(default_factory=list)

    @multimethod
    def visit(self, b: BaseBlock):
        self.merge_text()
        self.current_state.append(copy(b))

    @visit.register  # type: ignore
    def _(self, b: bk.Text):
        if b.name is None:
            self.current_text.append(b)
        else:
            self.merge_text()
            self.current_state.append(copy(b))

    @visit.register  # type: ignore
    def _(self, b: ContainerBlock):
        self.merge_text()

        if len(b.blocks) < b.report_minimum_blocks:
            msg = f"{b.__class__.__name__} has less than {b.report_minimum_blocks} objects"
            if self.is_finalized:
                raise ARError(msg)
            log.warning(msg)

        with self.fresh_state(b):
            self.in_collapsible_group = isinstance(b, Blocks) or (
                isinstance(b, bk.Group) and b.columns == 1
            )
            _ = b.traverse(self)
            self.merge_text()

    def merge_text(self):
        # log.debug("Merging text nodes")
        if self.current_text:
            t1: bk.Text  # noqa: F842

            if self.in_collapsible_group:
                new_text = "\n\n".join(t1.content for t1 in self.current_text)
                self.current_state.append(bk.Text(new_text))
            else:
                self.current_state.extend(copy(t1) for t1 in self.current_text)

            self.current_text = []

    @property
    def root(self) -> BaseBlock:
        return self.current_state[0]

    @contextmanager
    def fresh_state(self, b: ContainerBlock) -> Generator[None, None, None]:
        x = self.current_state
        self.current_state = []

        yield None

        # build a new instance of the container block
        b1 = copy(b)
        b1.blocks = self.current_state
        x.append(b1)
        self.current_state = x



from collections.abc import Mapping
from datetime import timedelta
from enum import Enum
from os import PathLike
from pathlib import Path
from typing import Any, NewType, Union

# Typedefs
# A JSON-serializable config object
SDict = dict[str, Any]
SSDict = dict[str, str]
SList = list[str]
# NOTE - mypy cannot handle recursive types like this currently. Will review in the future
JSON = Union[str, int, float, bool, None, Mapping[str, "JSON"], list["JSON"]]  # type: ignore
JDict = SDict  # should be JSON
JList = list[JSON]  # type: ignore
MIME = NewType("MIME", str)
URL = NewType("URL", str)
HTML = NewType("HTML", str)
NPath = Union[Path, PathLike, str]
Hash = NewType("Hash", str)
EnumType = int  # alias for enum values

# Constants
# NOTE - PKL_MIMETYPE and ARROW_MIMETYPE are custom mimetypes
PKL_MIMETYPE = MIME("application/vnd.pickle+binary")
ARROW_MIMETYPE = MIME("application/vnd.apache.arrow+binary")
ARROW_EXT = ".arrow"
TD_1_HOUR = timedelta(hours=1)
TD_1_DAY = timedelta(days=1)
SECS_1_HOUR: int = int(TD_1_HOUR.total_seconds())
SECS_1_WEEK: int = int(timedelta(weeks=1).total_seconds())
SIZE_1_MB: int = 1024 * 1024


class StrEnum(str, Enum):
    # TODO - replace with StrEnum in py3.11 stdlib
    def __str__(self):
        return str(self.value)



"""
Arakawa helper functions to improve the Arakawa UX in IPython notebooks
"""

from __future__ import annotations

from contextlib import suppress
from typing import TYPE_CHECKING

from arakawa.utils import display_msg

from .environment import get_environment
from .exceptions import BlocksNotFoundError, NotebookParityError

if TYPE_CHECKING:
    from arakawa.blocks import BaseBlock


def output_cell_to_block(cell: dict, ipython_output_cache: dict) -> BaseBlock | None:
    """Convert a IPython notebook output cell to An Arakawa Block"""
    from arakawa.blocks import wrap_block

    # Get the output object from the IPython output cache
    cell_output_object = ipython_output_cache.get(cell["execution_count"], None)

    # If there's no corresponding output object, skip
    if cell_output_object is not None:
        with suppress(Exception):
            return wrap_block(cell_output_object)

    return None


def check_notebook_cache_parity(
    notebook_json: dict, ipython_input_cache: list
) -> tuple[bool, list[int]]:
    """Check that the IPython output cache is in sync with the saved notebook"""
    is_dirty = False
    dirty_cells: list[int] = []

    # inline !bang commands (get_ipython().system), %line magics, and %%cell magics are not cached
    # exclude these from conversion
    ignored_cell_functions = [
        "get_ipython().system",
        "get_ipython().run_line_magic",
        "get_ipython().run_cell_magic",
    ]

    # broad check: check the execution count is the same
    execution_counts = [
        cell.get("execution_count", 0) or 0 for cell in notebook_json["cells"]
    ]

    latest_cell_execution_count = max(execution_counts, default=0)

    # -2 to account for zero-based indexing and the invoking cell not being saved
    latest_cache_execution_count = len(ipython_input_cache) - 2
    if latest_cache_execution_count != latest_cell_execution_count:
        is_dirty = True
        return is_dirty, dirty_cells

    # narrow check: check the cell source is the same for executed cells
    for cell in notebook_json["cells"]:
        cell_execution_count = cell.get("execution_count", None)
        if (
            cell["cell_type"] == "code"
            and cell_execution_count
            and cell_execution_count < len(ipython_input_cache)
        ):
            input_cache_source = ipython_input_cache[cell_execution_count].strip()

            # skip and mark cells containing ignored functions
            if any(
                ignored_function in input_cache_source
                for ignored_function in ignored_cell_functions
            ):
                cell["contains_ignored_functions"] = True
            # dirty because input has changed between execution and save.
            elif "".join(cell["source"]).strip() != input_cache_source:
                is_dirty = True
                dirty_cells.append(cell_execution_count)

    return is_dirty, dirty_cells


def cells_to_blocks(
    opt_out: bool = True, show_code: bool = False, show_markdown: bool = True
) -> list[BaseBlock]:
    """Convert IPython notebook cells to a list of Arakawa Blocks

    Recognized cell tags:
        - `ar-exclude` - Exclude this cell (when opt_out=True)
        - `ar-include` - Include this cell (when opt_out=False)
        - `ar-show-code` - Show the input code for this cell
        - `ar-show-markdown` - Show the markdown for this cell

    ..note:: IPython output caching must be enabled for this function to work. It is enabled by default.
    """
    environment = get_environment()
    if not environment.is_notebook_environment:
        raise ValueError("This function can only be used in a notebook environment")

    ip = environment.get_ipython()
    user_ns = ip.user_ns
    ipython_output_cache = user_ns["_oh"]
    ipython_input_cache = user_ns["_ih"]

    notebook_json = environment.get_notebook_json()
    # TODO: debug message for Colab, remove after testing

    notebook_is_dirty, dirty_cells = check_notebook_cache_parity(
        notebook_json, ipython_input_cache
    )

    if notebook_is_dirty:
        notebook_parity_message = "Please ensure all cells in the notebook have been executed and saved before running the conversion."

        if dirty_cells:
            notebook_parity_message += f"""

The following cells have not been executed and saved: {', '.join(map(str, dirty_cells))}"""

        raise NotebookParityError(notebook_parity_message)

    blocks = []

    for cell in notebook_json["cells"]:
        tags = cell["metadata"].get("tags", [])

        if (opt_out and "ar-exclude" not in tags) or (
            not opt_out and "ar-include" in tags
        ):
            if (cell["cell_type"] == "markdown" and cell.get("source")) and (
                show_markdown or "ar-show-markdown" in tags
            ):
                from arakawa.blocks.text import Text

                markdown_block: BaseBlock = Text("".join(cell["source"]))
                blocks.append(markdown_block)
            elif cell["cell_type"] == "code" and not cell.get(
                "contains_ignored_functions", False
            ):
                if "ar-show-code" in tags or show_code:
                    from arakawa.blocks.text import Code

                    code_block: BaseBlock = Code("".join(cell["source"]))
                    blocks.append(code_block)

                output_block = output_cell_to_block(cell, ipython_output_cache)

                if output_block:
                    blocks.append(output_block)
                elif "ar-include" in tags:
                    display_msg(
                        f'Cell output of type {type(ipython_output_cache.get(cell["execution_count"]))} not supported. Skipping.',
                    )

    if not blocks:
        raise BlocksNotFoundError("No blocks found.")

    display_msg("Notebook converted to blocks.")

    return blocks



from __future__ import annotations

from .base import BlockId, DataBlock


class ControlBlock(DataBlock):
    """
    Abstract block for all the control blocks.
    """


class BaseDateTimeBlock(ControlBlock):
    """
    Abstract block for DateTime, Date and Time blocks.
    """

    def __init__(
        self,
        name: BlockId,
        help_: str | None = None,
        initial: str | None = None,
        label: str | None = None,
        required: bool | None = None,
        validation: str | None = None,
    ):
        """
        Args:
            name (BlockId): A name.
            help_ (str | None, optional): A help text. Defaults to None.
            initial (str | None, optional): An initial value. Defaults to None.
            label (str | None, optional): A label. Defaults to None.
            required (bool | None, optional): Whether it's required or not. Defaults to None.
            validation (str | None, optional): A formkit validation in addition to required. Defaults to None.
        """
        super().__init__(
            name=name,
            label=label,
            initial=initial,
            required=required,
            help=help_,
            validation=validation,
        )


class DateTimeField(BaseDateTimeBlock):
    """
    DateTimeField allows you to add a datetime type input.
    """

    _tag = "DateTime"


class DateField(BaseDateTimeBlock):
    """
    DateField allows you to add a date type input.
    """

    _tag = "Date"


class TimeField(BaseDateTimeBlock):
    """
    TimeField allows you to add a time type input.
    """

    _tag = "Time"


class FileField(ControlBlock):
    """
    FileField allows you to add a file type input.
    """

    _tag = "File"

    def __init__(
        self,
        name: BlockId,
        accept: str | None = None,
        help_: str | None = None,
        label: str | None = None,
        required: bool | None = None,
        validation: str | None = None,
    ):
        """
        Args:
            name (BlockId): A name.
            help_ (str | None, optional): A help text. Defaults to None.
            label (str | None, optional): A label. Defaults to None.
            required (bool | None, optional): Whether it's required or not. Defaults to None.
            validation (str | None, optional): A formkit validation in addition to required. Defaults to None.
            accept (str | None, optional): An accept value. Defaults to None.
        """
        super().__init__(
            name=name,
            label=label,
            required=required,
            help=help_,
            validation=validation,
            accept=accept,
        )


class MultiChoiceField(ControlBlock):
    """
    MultiChoiceField allows you to have a multiple select type input.
    """

    _tag = "MultiChoice"

    def __init__(
        self,
        name: BlockId,
        initial: list[str],
        options: list[str],
        help_: str | None = None,
        label: str | None = None,
        required: bool | None = None,
        validation: str | None = None,
    ):
        """
        Args:
            name (BlockId): A name.
            initial (list[str]): An initial value.
            options (list[str]): Options.
            help_ (str | None, optional): A help text. Defaults to None.
            label (str | None, optional): A label. Defaults to None.
            required (bool | None, optional): Whether it's required or not. Defaults to None.
            validation (str | None, optional): A formkit validation in addition to required. Defaults to None.
        """
        super().__init__(
            initial=initial,
            options=options,
            name=name,
            label=label,
            required=required,
            help=help_,
            validation=validation,
        )


class NumberBox(ControlBlock):
    """
    NumberBox allows you to add a number type input.
    """

    _tag = "NumberBox"

    def __init__(
        self,
        name: BlockId,
        help_: str | None = None,
        initial: int | float | None = None,
        label: str | None = None,
        required: bool | None = None,
        validation: str | None = None,
    ):
        """
        Args:
            name (BlockId): A name.
            help_ (str | None, optional): A help text. Defaults to None.
            initial (int | float | None, optional): An initial value. Defaults to None.
            label (str | None, optional): A label. Defaults to None.
            required (bool | None, optional): Whether it's required or not. Defaults to None.
            validation (str | None, optional): A formkit validation in addition to required. Defaults to None.
        """
        super().__init__(
            name=name,
            label=label,
            initial=initial,
            required=required,
            help=help_,
            validation=validation,
        )


class RangeField(ControlBlock):
    """
    RangeField allows you to add a range type input.
    """

    _tag = "Range"

    def __init__(
        self,
        name: BlockId,
        min_: int | float,
        max_: int | float,
        step: int | float,
        help_: str | None = None,
        initial: int | float | None = None,
        label: str | None = None,
        validation: str | None = None,
    ):
        """
        Args:
            name (BlockId): Name.
            min_ (int | float): A min value.
            max_ (int | float): A max value.
            step (int | float): A step value.
            help_ (str | None, optional): A help text. Defaults to None.
            initial (int | float | None, optional): An initial value. Defaults to None.
            label (str | None, optional): A label. Defaults to None.
            validation (str | None, optional): A formkit validation in addition to required. Defaults to None.
        """
        initial = initial or min_
        super().__init__(
            min=min_,
            max=max_,
            step=step,
            name=name,
            label=label,
            initial=initial,
            help=help_,
            validation=validation,
        )


class ChoiceField(ControlBlock):
    """
    ChoiceField allows you to add a select type input.
    """

    _tag = "Choice"

    def __init__(
        self,
        name: BlockId,
        options: list[str],
        help_: str | None = None,
        initial: str | None = None,
        label: str | None = None,
        required: bool | None = None,
        validation: str | None = None,
    ):
        """
        Args:
            name (BlockId): A name.
            options (list[str]): Options.
            help_ (str | None, optional): A help text. Defaults to None.
            initial (str | None, optional): An initial value. Defaults to None.
            label (str | None, optional): A label. Defaults to None.
            required (bool | None, optional): Whether it's required or not. Defaults to None.
            validation (str | None, optional): A formkit validation in addition to required. Defaults to None.
        """
        super().__init__(
            options=options,
            name=name,
            label=label,
            initial=initial,
            required=required,
            help=help_,
            validation=validation,
        )


class SwitchField(ControlBlock):
    """
    SwitchField allows you to add a checkbox type input.
    """

    _tag = "Switch"

    def __init__(
        self,
        name: BlockId,
        help_: str | None = None,
        initial: bool | None = None,
        label: str | None = None,
        validation: str | None = None,
    ):
        """
        Args:
            name (BlockId): A name.
            help_ (str | None, optional): A help text. Defaults to None.
            initial (bool | None, optional): An initial value. Defaults to None.
            label (str | None, optional): A label. Defaults to None.
            validation (str | None, optional): A formkit validation in addition to required. Defaults to None.
        """
        super().__init__(
            name=name,
            label=label,
            initial=initial,
            help=help_,
            validation=validation,
        )


class TagsField(ControlBlock):
    """
    TagsField allows you to add a multi select type input along with the free form input.
    """

    _tag = "Tags"

    def __init__(
        self,
        name: BlockId,
        initial: list[str],
        help_: str | None = None,
        label: str | None = None,
        required: bool | None = None,
        validation: str | None = None,
    ):
        """
        Args:
            name (BlockId): A name.
            initial (list[str]): An initial value.
            help_ (str | None, optional): A help text. Defaults to None.
            label (str | None, optional): A label. Defaults to None.
            required (bool | None, optional): Whether it's required or not. Defaults to None.
            validation (str | None, optional): A formkit validation in addition to required. Defaults to None.
        """
        super().__init__(
            name=name,
            label=label,
            initial=initial,
            required=required,
            help=help_,
            validation=validation,
        )


class BaseTextField(ControlBlock):
    def __init__(
        self,
        name: BlockId,
        help_: str | None = None,
        initial: str | None = None,
        label: str | None = None,
        required: bool | None = None,
        validation: str | None = None,
    ):
        """
        Args:
            name (BlockId): A name.
            help_ (str | None, optional): A help text. Defaults to None.
            initial (str | None, optional): An initial value. Defaults to None.
            label (str | None, optional): A label. Defaults to None.
            required (bool | None, optional): Whether it's required or not. Defaults to None.
            validation (str | None, optional): A formkit validation in addition to required. Defaults to None.
        """
        super().__init__(
            name=name,
            label=label,
            initial=initial,
            required=required,
            validation=validation,
            help=help_,
        )


class TextBox(BaseTextField):
    """
    TextBox allows you to add a text type input.
    """

    _tag = "TextBox"


class URLField(BaseTextField):
    """
    URLField allows you to add a url type input.
    """

    _tag = "URL"


class EmailField(BaseTextField):
    """
    EmailField allows you to add an email type input.
    """

    _tag = "Email"


class SearchField(BaseTextField):
    """
    SearchField allows you to add a search type input.
    """

    _tag = "Search"


class TelephoneField(BaseTextField):
    """
    TelephoneField allows you to add a tel type input.
    """

    _tag = "Telephone"


class PasswordField(BaseTextField):
    """
    PasswordField allows you to add a password type input.
    """

    _tag = "Password"


class TextareaField(BaseTextField):
    """
    TextareaField allows you to add a search type input.
    """

    _tag = "Textarea"


class HiddenField(ControlBlock):
    """
    HiddenField allows you to add a hidden type input.
    """

    _tag = "Hidden"

    def __init__(
        self,
        name: BlockId,
        initial: str,
    ):
        """
        Args:
            name (BlockId): A name.
            initial (str): An initial value.
        """
        super().__init__(
            name=name,
            initial=initial,
        )


class ColorField(BaseTextField):
    """
    ColorField allows you to add a color picker.
    """

    _tag = "Color"



import datetime
from numbers import Number
from typing import Any

import numpy as np
import pandas as pd
from packaging.specifiers import SpecifierSet
from packaging.version import Version

PD_VERSION = Version(pd.__version__)
PD_1_3_GREATER = SpecifierSet(">=1.3.0")


def convert_axis(df: pd.DataFrame):
    """flatten both columns and indexes"""

    # flatten hierarchical columns and convert to strings
    if df.columns.nlevels > 1:
        df.columns = ["/".join(a) for a in df.columns.to_flat_index()]

    df.columns = df.columns.astype("string")

    # flatten/reset indexes
    if isinstance(df.index, pd.RangeIndex):
        pass  # allow RangeIndex - reset numbers?
    elif isinstance(df.index, np.integer):
        df.reset_index(inplace=True, drop=True)
    else:
        # reset if any other index type, e.g. MultiIndex, custom Index
        # all new column dtypes will converted in latter functions
        df.reset_index(inplace=True)


def downcast_numbers(data: pd.DataFrame):
    """Downcast numerics"""

    def downcast_ints(ser: pd.Series) -> pd.Series:
        try:
            ser = pd.to_numeric(ser, downcast="signed")
            ser = pd.to_numeric(ser, downcast="unsigned")
        except Exception:
            pass  # catch failure on Int64Dtype
        return ser

    # A result of downcast(timedelta64[ns]) is int <ns> and hard to understand.
    # e.g.) 0 days 00:54:38.777572 -> 3278777572000 [ns]
    df_num = data.select_dtypes("integer", exclude=["timedelta"])  # , pd.Int64Dtype])
    data[df_num.columns] = df_num.apply(downcast_ints)

    def downcast_floats(ser: pd.Series) -> pd.Series:
        return pd.to_numeric(ser, downcast="float", errors="ignore")

    # float downcasting currently disabled - alters values (both float64 and int64) and rounds to 'inf' instead of erroring
    # see https://github.com/pandas-dev/pandas/issues/19729
    # df_num = data.select_dtypes("floating")
    # data[df_num.columns] = df_num.apply(downcast_floats)


def timedelta_to_str(df: pd.DataFrame):
    """
    convert timedelta to str
    NOTE - only until arrow.js supports Duration type
    """
    df_td = df.select_dtypes("timedelta")
    df[df_td.columns] = np.where(pd.isnull(df_td), pd.NA, df_td.astype("string"))


def parse_categories(data: pd.DataFrame):
    """Detect and converts categories"""

    def criteria(ser: pd.Series) -> bool:
        """Decides whether to convert into categorical"""
        nunique: int = ser.nunique()

        if nunique <= 20 and (nunique != ser.size):
            # few unique values => make it a category regardless of the proportion
            return True

        prop_unique = (nunique + 1) / (ser.size + 1)  # + 1 for nan

        # a lot of redundant information => categories are more compact
        return prop_unique <= 0.05

    def try_to_category(ser: pd.Series) -> pd.Series:
        return ser.astype("category") if criteria(ser) else ser

    potential_cats = data.select_dtypes(["string", "object"])
    data[potential_cats.columns] = potential_cats.apply(try_to_category)


def obj_to_str(df: pd.DataFrame):
    """Converts remaining objects columns to str"""
    # convert objects to str / NA
    df_str = df.select_dtypes("object")
    df[df_str.columns] = df_str.astype("string")

    # convert categorical values (as strings if object)
    def to_str_cat_vals(x: pd.Series) -> pd.Series:
        if x.cat.categories.dtype == np.dtype("object"):
            x = x.cat.set_categories(x.cat.categories.astype("string"))

        return x

    df_cat = df.select_dtypes("category")
    df[df_cat.columns] = df_cat.apply(to_str_cat_vals)


def bipartite_to_bool(df: pd.DataFrame):
    # This was removed from our processing steps as some users required the numerical representation of binary columns.
    """Converts biperatite numeric {0, 1} columns to bool."""
    # Get names of numeric columns with only 2 unique values.
    df_num = df.select_dtypes("integer", exclude=["timedelta"])
    bipartite_columns = df_num.columns[df_num.dropna().nunique() == 2]

    for column in bipartite_columns:
        series = df[column]

        # Only apply the type change to {0, 1} columns
        val_range = series.min(), series.max()
        val_min, val_max = val_range[0], val_range[1]

        if val_min == 0 and val_max == 1:
            df[column] = df[column].astype(bool)


def str_to_arrow_str(df: pd.DataFrame):
    """Use the memory-efficient pyarrow string dtype (pandas >= 1.3 only)"""
    # convert objects to str / NA
    df_str = df.select_dtypes("string")
    df[df_str.columns] = df_str.astype("string[pyarrow]")


def process_df(df: pd.DataFrame, copy: bool = False) -> pd.DataFrame:
    """
    Processing steps needed before writing / after reading
    We only modify the dataframe to optimise size,
    rather than convert/infer types, e.g. no longer parsing dates from strings

    NOTE - this mutates the dataframe by default but returns it - use the returned copy!
    """
    if copy:
        df = df.copy(deep=True)

    convert_axis(df)

    # convert timedelta
    timedelta_to_str(df)

    # NOTE - pandas >= 1.3 handles downcasting of nullable values correctly
    df = df.convert_dtypes()
    downcast_numbers(df)

    # save timedeltas cols (unneeded whilst timedelta_to_str used)
    # td_col = df.select_dtypes("timedelta")
    # df[td_col.columns] = td_col
    obj_to_str(df)
    parse_categories(df)

    # convert all strings to use the arrow dtype
    str_to_arrow_str(df)

    return df


def to_df(value: Any) -> pd.DataFrame:
    """
    Converts a python object, i.e. a app's output, to a dataframe
    NOTE - this returns a new DF each time
    """
    if value is None:
        # This return the empty dataframe, which atm is the same as
        # the empty file object in the CAS.
        # However this is not ensured as pyarrow changes
        return pd.DataFrame()

    if isinstance(value, pd.DataFrame):
        return value.copy(deep=True)

    if isinstance(value, (pd.Series, pd.Index)):
        if value.name is not None:
            return pd.DataFrame(value)

        return pd.DataFrame({"Result": value})

    if isinstance(value, (Number, str, bool, datetime.datetime, datetime.timedelta)):
        return pd.DataFrame({"Result": value}, index=[0])

    if isinstance(value, np.ndarray):
        try:
            out_df = pd.DataFrame(value)
        except ValueError:
            squeezed = np.squeeze(value)
            if squeezed.shape == ():
                # must be a scalar
                out_df = pd.DataFrame({"Result": squeezed}, index=[0])
            else:
                out_df = pd.DataFrame(squeezed)

        if out_df.columns.tolist() == [0]:
            out_df.columns = ["Result"]

        return out_df

    raise ValueError("Must return a primitive, pd.DataFrame, pd.Series or numpy array.")


TRUNCATE_CELLS = 10000
TRUNCATE_ROWS = 1000


def truncate_dataframe(
    df: pd.DataFrame, max_rows: int = TRUNCATE_ROWS, max_cells: int = TRUNCATE_CELLS
) -> pd.DataFrame:
    """Truncate a pandas dataframe if needed"""
    rows, cols = df.shape
    # determine max rows to truncate df to based on max cells and df cols
    cols = cols or 1  # handle empty df
    max_rows = min(max_rows, int(max_cells / cols))
    # return non-truncated preview if df smaller than max rows allowed
    if rows <= max_rows:
        return df
    # truncate df to fit max cells
    if not isinstance(df.index, pd.RangeIndex):
        raise ValueError("Dataframe has unsupported index type")
    return df.truncate(before=0, after=max_rows - 1, copy=False)
